/*----------------------------------------------------------------------
  File    : client.c
  Contents: hamster control functions, client side
            sends commands to stdout, receives from stdin
  Author  : Christian Borgelt
  History : 21.10.1997 file created
            02.11.1997 function hms_move simplified
            15.01.1998 initialization of random number generator added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include "hamster.h"

/*----------------------------------------------------------------------
  Protocol for Pipe Communication
------------------------------------------------------------------------
request        meaning         reply
c              create hamster  c <id> <x> <y> <dir> <look> <corn> <load>
                               c -1      on error
d <id>         delete hamster  d <id>
m <id>         move hamster    m <id> <x> <y> <look> <corn>
t <id> <turn>  turn hamster    t <id> <dir> <look>
l <id> <n>     (un)load corn   l <id> <corn> <load>                   */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- error codes --- */
#define E_NONE       0          /* no error */
#define E_NOMEM    (-1)         /* not enough memory */
#define E_PREAD    (-2)         /* file read failed */
#define E_PWRITE   (-3)         /* file write failed */
#define E_UNKNOWN  (-4)         /* unknown error */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
struct hamster {                /* --- a hamster --- */
  int id;                       /* hamster identifier */
  int x, y;                     /* relative position in maze */
  int dir;                      /* direction (line of sight) */
  int look;                     /* outlook in current direction */
  int corn;                     /* amount of corn on current field */
  int load;                     /* amount of corn in cheeks */
};                              /* (type HAMSTER) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_PREAD    -2 */  "read error on pipe",
  /* E_PWRITE   -3 */  "write error on pipe",
  /* E_UNKNOWN  -4 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static char *prgname;           /* program name for error messages */

/*----------------------------------------------------------------------
  Error Function
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- error function */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check the error code and */
  msg = errmsgs[-code];         /* get the error message */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  fprintf(stderr, "\n%s: ", prgname);
  va_start(args, code);         /* get variable arguments */
  vfprintf(stderr, msg, args);  /* print the error message */
  va_end(args);                 /* end argument evaluation */
  exit(code);                   /* abort the programm */
}  /* error() */

/*----------------------------------------------------------------------
  Internal Hamster Functions
----------------------------------------------------------------------*/

static HAMSTER* hms_create (void)
{                               /* --- create a hamster ('c') */
  int     id;                   /* hamster identifier */
  HAMSTER *hms;                 /* created hamster */

  if ((printf("c\n") < 0) || fflush(stdout))
    error(E_PWRITE);            /* send create message */
  if (scanf("c %d", &id) != 1)  /* receive and */
    error(E_PREAD);             /* scan answer */
  if (id < 0) return NULL;      /* if create failed, abort */
  hms = (HAMSTER*)malloc(sizeof(HAMSTER));
  if (!hms) error(E_NOMEM);     /* allocate memory */
  if ((scanf(" %d %d %d", &hms->x,    &hms->y,    &hms->dir)  != 3)
  ||  (scanf(" %d %d %d", &hms->look, &hms->corn, &hms->load) != 3))
    error(E_PREAD);             /* receive hamster data */
  hms->id = id;                 /* and set hamster id */
  return hms;                   /* return created hamster */
}  /* hms_create() */

/*--------------------------------------------------------------------*/

static void hms_delete (HAMSTER *hms)
{                               /* --- delete a hamster ('d') */
  int id;                       /* hamster identifier, buffer */

  if ((printf("d %d\n", hms->id) < 0) || fflush(stdout))
    error(E_PWRITE);            /* send delete message */
  if ((scanf(" d %d", &id) != 1) || (id != hms->id))
    error(E_PREAD);             /* scan the answer */
  free(hms);                    /* deallocate memory */
}  /* hms_delete() */

/*----------------------------------------------------------------------
  External Hamster Functions
----------------------------------------------------------------------*/

void hms_pos (HAMSTER *hms, int *x, int *y)
{                               /* --- get hamster position */
  *x = hms->x; *y = hms->y;     /* return relative coordinates */
}  /* hms_pos() */

/*--------------------------------------------------------------------*/

int hms_dir (HAMSTER *hms)
{                               /* --- get hamster direction */
  return hms->dir;              /* return current direction */
}  /* hms_dir() */

/*--------------------------------------------------------------------*/

int hms_look (HAMSTER *hms)
{                               /* --- take a look ahead */
  return hms->look;             /* return outlook in current dir. */
}  /* hms_look() */

/*--------------------------------------------------------------------*/

int hms_corn (HAMSTER *hms)
{                               /* --- check for corn */
  return hms->corn;             /* return amount of corn on field */
}  /* hms_corn() */

/*--------------------------------------------------------------------*/

int hms_load (HAMSTER *hms)
{                               /* --- get hamster load */
  return hms->load;             /* return amount of corn in cheeks */
}  /* hms_load() */

/*--------------------------------------------------------------------*/

int hms_move (HAMSTER *hms)
{                               /* --- move hamster one field ('m') */
  int id;                       /* hamster identifier */
  int x, y;                     /* old hamster position */

  x = hms->x, y = hms->y;       /* note current hamster position */
  if ((printf("m %d\n", hms->id) < 0) || fflush(stdout))
    error(E_PWRITE);            /* send move message */
  if ((scanf(" m %d", &id) != 1) || (id != hms->id)
  ||  (scanf(" %d %d", &hms->x,    &hms->y)    != 2)
  ||  (scanf(" %d %d", &hms->look, &hms->corn) != 2))
    error(E_PREAD);             /* get new hamster data */
  return ((hms->x == x) && (hms->y == y)) ? -1 : 0;
}  /* hms_move() */

/*--------------------------------------------------------------------*/

void hms_turn (HAMSTER *hms, int turn)
{                               /* --- turn hamster 90 degrees ('t') */
  int id;                       /* hamster identifier */

  if ((printf("t %d %d\n", hms->id, turn) < 0) || fflush(stdout))
    error(E_PWRITE);            /* send turn message */
  if ((scanf(" t %d", &id) != 1) || (id != hms->id)
  ||  (scanf("%d %d", &hms->dir, &hms->look) != 2))
    error(E_PREAD);             /* scan the answer */
}  /* hms_turn() */

/*--------------------------------------------------------------------*/

int hms_take (HAMSTER *hms, int amount)
{                               /* --- take/drop some corn ('l') */
  int id;                       /* hamster identifier */
  int load;                     /* buffer for corn load */

  load = hms->load;             /* note current load */
  if ((printf("l %d %d\n", hms->id, amount) < 0) || fflush(stdout))
    error(E_PWRITE);            /* send take/drop message */
  if ((scanf(" l %d", &id) != 1) || (id != hms->id)
  ||  (scanf(" %d %d", &hms->corn, &hms->load) != 2))
    error(E_PREAD);             /* scan the answer */
  return hms->load -load;       /* return amount taken/dropped */
}  /* hms_take() */

/*--------------------------------------------------------------------*/

int hms_drop (HAMSTER *hms, int amount)
{                               /* --- drop/take some corn */
  return -hms_take(hms, -amount);
}  /* hms_drop() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  HAMSTER *hms;                 /* controlled hamster */

  prgname = argv[0];            /* get program name for error msgs. */
  srand((unsigned)time(NULL));  /* initialize random numbers */
  hms = hms_create();           /* create a hamster */
  if (!hms) error(E_UNKNOWN);   /* check for success */
  hms_ctrl(hms);                /* call control function */
  hms_delete(hms);              /* delete hamster */
  return 0;                     /* terminate program */
}  /* main() */
