/*----------------------------------------------------------------------
  File    : maze.c
  Contents: maze management
  Author  : Christian Borgelt
  History : 1997.10.14 file created
            1997.10.15 first version completed
            1997.10.31 minor improvements
            1997.11.02 use of wall probability corrected
            1998.01.04 bug in maze initialization fixed
            2013.08.09 adapted to higher compiler warning level
----------------------------------------------------------------------*/
#include <stdlib.h>
#include "maze.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define DIRMASK  0x03           /* mask for directions */
#define VISITED  0x04           /* flag for 'field visited' */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- direction information --- */
  short dx, dy;                 /* field offsets */
  short wall;                   /* corresponding wall flag */
  short opp;                    /* opposite direction */
} DIRINFO;                      /* (direction information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const DIRINFO dinfo[] = {/* direction information */
  {  1,  0, MZ_EAST,  2 },      /* 0: east,  opposite is 2: west  */
  {  0,  1, MZ_NORTH, 3 },      /* 1: north, opposite is 3: south */
  { -1,  0, MZ_WEST,  0 },      /* 2: west,  opposite is 0: east  */
  {  0, -1, MZ_SOUTH, 1 } };    /* 3: south, opposite is 1: north */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

MAZE* mz_create (int xext, int yext)
{                               /* --- create a maze */
  MAZE  *maze;                  /* created maze */
  int   x;                      /* loop variable */
  short **col;                  /* to traverse column vector */

  if (xext <= 0) xext = 1;      /* check maze extensions */
  if (yext <= 0) yext = 1;      /* (both must be at least one) */
  maze = (MAZE*)calloc(1, sizeof(MAZE)+(size_t)(xext-1)*sizeof(short*));
  if (!maze) return NULL;       /* allocate maze body */
  maze->xext = xext;            /* and initialize fields */
  maze->yext = yext;
  maze->xpos = maze->ypos = -1; /* clear start position */
  col = maze->map;              /* traverse column vector */
  for (x = xext; --x >= 0; col++) {
    *col = (short*)malloc((size_t)yext *sizeof(short));
    if (!*col) { mz_delete(maze); return NULL; }
  }                             /* allocate maze columns */
  mz_init(maze, 0, 0.0F, 0, 0); /* and initialize maze */
  return maze;                  /* return created maze */
}  /* mz_create() */

/*--------------------------------------------------------------------*/

void mz_delete (MAZE *maze)
{                               /* --- delete a maze */
  int   x;                      /* loop variable */
  short **col;                  /* to traverse column vector */

  col = maze->map;              /* traverse maze columns */
  for (x = maze->xext; --x >= 0; col++)
    if (*col) free(*col);       /* delete maze columns */
  free(maze);                   /* and maze body */
}  /* mz_delete() */

/*--------------------------------------------------------------------*/

void mz_init (MAZE *maze, double (*randfn)(void),
              float wallprob, int total, int maxcnt)
{                               /* --- initialize a maze */
  short i;                      /* loop variable, buffer */
  int   x, y;                   /* loop variables, field indices */
  short **col;                  /* to traverse column vector */
  short *src, *dst;             /* to traverse columns, buffers */
  int   plen, idx;              /* length of path, field index */
  short dir;                    /* current direction */
  short cand[4];                /* direction candidates */
  short build;                  /* whether to build walls */
  int   cnt;                    /* number of candidates/fields  */
  short mincnt;                 /* minimal number of items per field */
  const DIRINFO *cdi;           /* current direction information */

  /* --- clear maze --- */
  for (col = maze->map, x = maze->xext; --x >= 0; )
    for (src = *col++, y = maze->yext; --y >= 0; )
      *src++ = 0;               /* clear all fields of the maze */
  col = maze->map;              /* traverse column vector */
  src = col[0];                 /* traverse first */
  dst = col[maze->xext-1];      /* and last column */
  for (y = maze->yext; --y >= 0; ) {
    *dst++ |= MZ_EAST; *src++ |= MZ_WEST;
  }                             /* set east and west border */
  i = (short)(maze->yext-1);    /* get index of northest row */
  for (x = maze->xext; --x >= 0; col++) {
    (*col)[i] |= MZ_NORTH; (*col)[0] |= MZ_SOUTH;
  }                             /* set north and south border */

  /* --- randomly place walls --- */
  if (randfn && (wallprob > 0.0)) {
    x   = (maze->xext -1) >> 1; /* compute coordinates */
    y   = (maze->yext -1) >> 1; /* of center field and */
    src = maze->map[x] +y;      /* initialize its direction */
    *src = (short)(*src | ((short)(randfn() *4) & DIRMASK) | VISITED);
    plen = 1; build = 0;        /* path contains the center */
    while (plen > 0) {          /* while path is not empty */
      cnt = 0;                  /* clear candidate counter */
      dir = *src & DIRMASK;     /* get direction to current field */
      for (i = 3; --i >= 0; ) { /* traverse remaining directions */
        dir = (dir +1) & DIRMASK;      /* compute next direction */
        cdi = dinfo +dir;       /* and get direction information */
        if (*src & cdi->wall)   /* if direction is blocked, */
          continue;             /* skip this direction */
        dst = maze->map[x +cdi->dx] +(y +cdi->dy);
        if (!(*dst & VISITED))  /* if field has not been visited, */
          cand[cnt++] = dir;    /* note direction candidate */
        else if (build          /* otherwise build a wall randomly */
        &&       ((*dst & DIRMASK) != cdi->opp)
        &&       (randfn() <= wallprob)) {
          *src |= cdi->wall;    /* build wall on current field */
          *dst |= dinfo[cdi->opp].wall;
        }                       /* build wall on destination field */
      }                         /* (for consistency) */
      if (cnt <= 0) {           /* if no candidates found, */
        dir = *src & DIRMASK;   /* get direction by which */
        x += dinfo[dir].dx;     /* current field was entered */
        y += dinfo[dir].dy;     /* and offsets to previous field */
        src = maze->map[x] +y;  /* retrace step (go back on path) */
        plen--; build = 0; }    /* reduce path length, clear flag */
      else {                    /* if at least one candidate found, */
        dir = (short)(randfn() *cnt);
        dir = cand[(dir >= cnt) ? cnt-1 : dir];
        x += dinfo[dir].dx;     /* choose direction to move in, */
        y += dinfo[dir].dy;     /* get offsets to new field */
        src = maze->map[x] +y;  /* and move to new field */
        *src |= dinfo[dir].opp | VISITED;
        plen++; build = 1;      /* note direction from which */
      }                         /* field was entered and */
    }                           /* increase path length */
  }  /* if (randfn && (wallprob > 0.0)) */

  /* --- clear markers and place items --- */
  if (maxcnt > MZ_ITEMS)        /* at most MZ_ITEMS items */
    maxcnt = MZ_ITEMS;          /* can be placed on one field */
  cnt = (int)maze->xext *maze->yext -1;
  if (total > cnt *maxcnt)      /* check total number of items */
    total = cnt *maxcnt;        /* against the maze capacity */
  if (!randfn)                  /* compute average number of items */
    maxcnt = (cnt <= 0) ? 0 : (int)(total /cnt +((total %cnt) ? 1 : 0));
  col = maze->map;              /* traverse column vector */
  for (x = maze->xext; --x >= 0; ) {
    dst = *col++;               /* traverse columns */
    for (y = maze->yext; --y >= 0; dst++, cnt--) {
      if (total < maxcnt)       /* if less than 'maxcnt' items left, */
        maxcnt = (int)total;    /* reduce upper limit to 'total' */
      idx    = total -(cnt-1) *maxcnt;  /* compute lower limit on */
      mincnt = (short)((idx < 1) ? 1 : idx); /* # items per field */
      if (!randfn               /* if no random number function given */
      ||  (mincnt >= maxcnt))   /* or number of items is fixed, */
        i = (short)maxcnt;      /* choose maximal number of items */
      else {                    /* if number of items is variable */
        i = (short)(mincnt +randfn() *(maxcnt-mincnt+1));
        if (i < mincnt) i = (short)mincnt;
        if (i > maxcnt) i = (short)maxcnt;
      }                         /* choose number of items randomly */
      *dst   = (*dst & MZ_WALLS) | i;
      total -= i;               /* place 'i' items on the field */
    }                           /* and reduce total number of items */
  }                             /* by the same amount */

  /* --- shuffle item counters of fields --- */
  if (randfn) {                 /* if random number function given */
    i   = 1;                    /* set first field flag */
    cnt = maze->xext *maze->yext;     /* compute number of fields */
    col = maze->map +maze->xext;      /* traverse column vector */
    for (x = maze->xext; --x >= 0; ) {
      dst = (*--col) +maze->yext;     /* traverse columns */
      for (y = maze->yext; --y >= 0; cnt--) {
        if (i) {                      /* if this is the first field, */
          dst--; i = 0; continue; }   /* skip start position buffer */
        idx = (int)(randfn() *cnt);   /* choose random field index */
        if ((idx < 0) || (idx >= cnt)) idx = 0;
        src = maze->map[idx /maze->yext] +(idx %maze->yext);
        dir  = *--dst & MZ_ITEMS;
        *dst = (short)((*dst & ~MZ_ITEMS) | (*src & MZ_ITEMS));
        *src = (short)((*src & ~MZ_ITEMS) | ( dir & MZ_ITEMS));
      }                         /* exchange the item counters */
    }                           /* of the randomly chosen field */
  }                             /* and the current field */

  /* --- choose start position --- */
  if (!randfn)                  /* if no random number function given,*/
    maze->xpos = maze->ypos = 0;/* start from south west corner */
  else {                        /* if random number function given */
    maze->xpos = (int)(randfn() *maze->xext);
    if ((maze->xpos < 0) || (maze->xpos >= maze->xext))
      maze->xpos = 0;           /* choose random x-coordinate */
    maze->ypos = (int)(randfn() *maze->yext);
    if ((maze->xpos < 0) || (maze->ypos >= maze->yext))
      maze->ypos = 0;           /* choose random y-coordinate */
  }                             /* of the start position */
  src = maze->map[maze->xext-1] +(maze->yext-1);
  dst = maze->map[maze->xpos]   + maze->ypos;
  *src = (short)((*dst &  MZ_ITEMS) | *src);    /* clear and mark */
  *dst = (short)((*dst & ~MZ_ITEMS) | MZ_HOME); /* start position */
}  /* mz_init() */

/*--------------------------------------------------------------------*/

void mz_setpos (MAZE *maze, int x, int y)
{                               /* --- set start position */
  short ox, oy;                 /* coordinates of old start position */
  short tmp;                    /* temporary buffer */

  ox = (short)maze->xpos; oy = (short)maze->ypos;
  if ((ox >= 0) && (oy >= 0))   /* if old position was valid, */
    maze->map[ox][oy] &= ~MZ_HOME;   /* clear home flag */
  if ((x < 0) || (x >= maze->xext)   /* if new position */
  ||  (y < 0) || (y >= maze->yext))  /* is invalid, */
    maze->xpos = maze->ypos = -1;    /* clear coordinates */
  else {                        /* if new position is valid */
    maze->xpos = x; maze->ypos = y;  /* note coordinates and */
    tmp = maze->map[x][y];           /* clear counter, set home flag */
    maze->map[x][y] = (tmp & ~MZ_ITEMS) | MZ_HOME;
    if ((ox >= 0) && (oy >= 0)) /* if old position was valid */
      maze->map[ox][oy] = (short)(maze->map[ox][oy] | (tmp & MZ_ITEMS));
  }                             /* transfer items to old field */
}  /* mz_setpos() */

/*--------------------------------------------------------------------*/

short mz_getfld (MAZE *maze, int x, int y)
{                               /* --- get field contents */
  if ((x < 0) || (x >= maze->xext)
  ||  (y < 0) || (y >= maze->yext))
    return 0;                   /* if field is not in maze, abort */
  return maze->map[x][y];       /* return field contents */
}  /* mz_getfld() */

/*--------------------------------------------------------------------*/

short mz_setfld (MAZE *maze, int x, int y, short mask, short cont)
{                               /* --- set field contents */
  short *fld;                   /* maze field to change */
  short chgd;                   /* changed walls */

  if ((x < 0) || (x >= maze->xext)
  ||  (y < 0) || (y >= maze->yext))
    return 0;                   /* if field is not in maze, abort */
  fld = maze->map[x] +y;        /* get maze field to change */
  chgd = (*fld ^ cont) & mask;  /* get change and set new contents */
  *fld = (short)((*fld & ~mask) | (cont & mask));
  if (!(chgd & MZ_WALLS))       /* if no walls were changed, */
    return *fld;                /* abort function */
  if      (x >= maze->xext-1) *fld              |= MZ_EAST;
  else if (chgd & MZ_EAST)    maze->map[x+1][y] ^= MZ_WEST;
  if      (x <= 0)            *fld              |= MZ_WEST;
  else if (chgd & MZ_WEST)    maze->map[x-1][y] ^= MZ_EAST;
  if      (y >= maze->yext-1) *fld              |= MZ_NORTH;
  else if (chgd & MZ_NORTH)   maze->map[x][y+1] ^= MZ_SOUTH;
  if      (y <= 0)            *fld              |= MZ_SOUTH;
  else if (chgd & MZ_SOUTH)   maze->map[x][y-1] ^= MZ_NORTH;
  return *fld;                  /* update wall flags */
}  /* mz_setfld() */

/*--------------------------------------------------------------------*/

int mz_save (MAZE *maze, FILE *file)
{                               /* --- save a maze */
  int   x, y;                   /* loop variables */
  short **col;                  /* to traverse column vector */

  fprintf(file, "%i %i ",  maze->xext, maze->yext);
  fprintf(file, "%i %i\n", maze->xpos, maze->ypos);
  for (y = maze->yext; --y >= 0; ) { /* traverse rows */
    col = maze->map;            /* traverse columns */
    for (x = maze->xext; --x > 0; col++)
      fprintf(file, "%04hx ", (*col)[y]);
    fprintf(file, "%04hx\n", (*col)[y]);
  }                             /* print field contents */
  return ferror(file) ? -1 : 0; /* return error status */
}  /* mz_save() */

/*--------------------------------------------------------------------*/

MAZE* mz_load (FILE *file)
{                               /* --- load a maze */
  MAZE  *maze;                  /* loaded maze */
  int   x, y, sx, sy;           /* loop variables, buffers */
  short **col;                  /* to traverse column vector */

  if ((fscanf(file, "%i %i ",  &x,  &y)  != 2)
  ||  (fscanf(file, "%i %i\n", &sx, &sy) != 2))
    return NULL;                /* read maze extensions and */
  maze = mz_create(x, y);       /* start position coordinates */
  if (!maze) return NULL;       /* create a maze */
  maze->xpos = sx; maze->ypos = sy;
  for (y = maze->yext; --y >= 0; ) { /* traverse rows */
    col = maze->map;            /* traverse columns */
    for (x = maze->xext; --x >= 0; col++) {
      if (fscanf(file, "%04hx", (short unsigned int*)((*col)+y)) != 1) {
        mz_delete(maze); return NULL; }
    }                           /* read maze field contents */
  }                             /* and store it */
  return maze;                  /* return loaded maze */
}  /* mz_load() */
