/*----------------------------------------------------------------------
  File    : xhamster.c
  Contents: main program for hamster (X11 version)
  Author  : Christian Borgelt
  History : 1997.10.21 file created
            1997.10.22 field drawing completed
            1997.10.26 dialog boxes programmed
            1997.10.27 dialog boxes completed, menu accelerators added
            1997.10.28 maze widget translations added
            1997.10.30 functions to load and save a maze implemented
            1997.10.31 minor improvements
            1997.11.01 child process and pipe communication added
            1997.11.02 child process and pipe communication debugged
            1997.11.04 keyboard focus changing added
            1997.11.06 event loops in icb_pipe and error fixed
            1997.11.09 bug in warning messages fixed
            1997.11.10 bug in function dcb_load (missing resize) fixed
            1997.12.21 some local variables removed
            1998.01.02 minor improvements
            1998.01.04 menu popup callback function removed
            1998.01.16 constant MAXHEAP defined as HMS_MAXCORN
            1998.01.17 counter for crashes against walls added
            1998.04.10 minor improvements
            1998.04.16 adapted to simplified menu definition
            1998.05.15 maze form widget removed
            1998.06.12 heap size display added, bug in offsets fixed
            1998.06.16 menu items Reload Maze.../Save Maze As... added
            1999.02.08 dialog boxes adapted to module psheet
            1999.03.08 '#include <sys/types.h>' added for Linux
            2000.01.12 adapted to module dialog
            2001.05.05 function 'resize' redesigned
            2013.08.09 adapted to higher compiler warning level
----------------------------------------------------------------------*/
#define _POSIX_SOURCE
#define select SELECT           /* just to fool HPUX's includes */
#define random RANDOM
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#undef select
#undef random
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "sprite.h"
#include "server.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define CCHAR    const char     /* abbreviation */
#define ULONG    unsigned long  /* ditto */
#define NONAME   "<noname>"     /* name for unsaved maze */

#define BDPENALTY   1000        /* penalty for program breakdown */

/* --- extensions --- */
#define HMS_MINXEXT    1        /* minimal x-extension of maze */
#define HMS_MINYEXT    1        /* minimal y-extension of maze */
#define MAXHEAP     HMS_MAXCORN /* maximal size of corn heap */
#define MINSPEED     0.1        /* minimal hamster speed */
#define MAXSPEED   100.0        /* maximal hamster speed */
#define FIELDWD       16        /* width of maze field (pixels) */
#define SPRITEWD      13        /* width of hamster sprite (pixels) */

/* --- menu definitions --- */
#define MENUCNT        3        /* number of menus (titles) */
#define ITEMCNT       12        /* number of menu items */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_POPEN      (-5)       /* cannot open pipe */
#define E_PREAD      (-6)       /* read error on pipe */
#define E_PWRITE     (-7)       /* write error on pipe */
#define E_PBROKEN    (-8)       /* broken pipe */
#define E_FORK       (-9)       /* cannot fork child process */
#define E_EXEC      (-10)       /* cannot execute program */
#define E_INIT      (-11)       /* initialization failed */
#define E_WIDGET    (-12)       /* widget creation failed */
#define E_COLOR     (-13)       /* color allocation failed */
#define E_GRAPH     (-14)       /* cannot create graphics context */
#define E_DIALOG    (-15)       /* cannot create dialog box widget */
#define E_UNKNOWN   (-16)       /* unkown error */

#define FN_MAX       (PATH_MAX +256)    /* max. length of file names */
#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- application specific data */
  int     xext, yext;           /* maze extensions */
  float   wallprob;             /* wall probability */
  int     total;                /* total corn in maze */
  int     maxheap;              /* maximal size of corn heap */
  float   speed;                /* hamster speed (fields/second) */
  int     corncnt;              /* total corn collected */
  int     movecnt;              /* number of moves made */
  int     crshcnt;              /* number of crashes against walls */
  int     score;                /* score achieved */
  char    fn_maze[FN_MAX];      /* file name of current maze */
  char    fn_hms [FN_MAX];      /* file name of executable */
} AppData;                      /* (application specific data) */

typedef struct {                /* --- hamster control block --- */
  HAMSTER *hamster;             /* hamster to control */
  SPRITE  *sprite;              /* sprite associated with hamster */
  char    reply[64];            /* reply to be sent to client */
} HMSCB;                        /* (hamster control block) */

typedef struct {                /* --- error message data --- */
  CCHAR   *name, *type;         /* name and type of the error message */
  CCHAR   *class;               /* resource class of error message */
  CCHAR   *dflt;                /* default message */
  int     argc;                 /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Function Prototypes
----------------------------------------------------------------------*/
static void reply    (Widget w, XtPointer client, XtPointer call);
static void result   (void);

/* --- actions --- */
static void load     (Widget w, XEvent *e, String *s, Cardinal *c);
static void save     (Widget w, XEvent *e, String *s, Cardinal *c);
static void select   (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit     (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw   (Widget w, XEvent *e, String *s, Cardinal *c);
static void size     (Widget w, XEvent *e, String *s, Cardinal *c);
static void random   (Widget w, XEvent *e, String *s, Cardinal *c);
static void clear    (Widget w, XEvent *e, String *s, Cardinal *c);
static void speed    (Widget w, XEvent *e, String *s, Cardinal *c);
static void exec     (Widget w, XEvent *e, String *s, Cardinal *c);
static void about    (Widget w, XEvent *e, String *s, Cardinal *c);
static void wall     (Widget w, XEvent *e, String *s, Cardinal *c);
static void heap     (Widget w, XEvent *e, String *s, Cardinal *c);
static void position (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xhamster.rsc"         /* fallback resources */
#include "hamster.xbm"          /* hamster sprite bitmap */

static Cardinal one = 1;        /* the number one (action parameter) */

static XtActionsRec actions[] = {
  { "load",     load     },     /* File    > Load Maze... */
  { "save",     save     },     /* File    > Save Maze... */
  { "select",   select   },     /* File    > Select Hamster... */
  { "quit",     quit     },     /* File    > Quit */
  { "redraw",   redraw   },     /* Actions > Redraw */
  { "size",     size     },     /* Actions > Set Maze Size... */
  { "random",   random   },     /* Actions > Randomize Maze... */
  { "clear",    clear    },     /* Actions > Clear Maze */
  { "speed",    speed    },     /* Actions > Set Hamster Speed... */
  { "exec",     exec     },     /* Actions > Start/Stop Hamster */
  { "about",    about    },     /* Help    > About Hamster */
  { "wall",     wall     },     /* toggle wall/remove walls */
  { "heap",     heap     },     /* change corn heap size */
  { "position", position },     /* set start position */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "Xhamster.Error",
                         "no error",                          0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "XHamster.Error",
                         "not enough memory",                 0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "XHamster.Warning",
                         "cannot open file:\n%s",             1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "XHamster.Warning",
                         "read error on file:\n%s",           1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "XHamster.Warning",
                         "write error on file:\n%s",          1 },
  /* E_POPEN    -5 */  { "warning", "popen",   "XHamster.Warning",
                         "cannot open pipe to child process", 0 },
  /* E_PREAD    -6 */  { "warning", "pread",   "XHamster.Warning",
                         "read error on pipe",                0 },
  /* E_PWRITE   -7 */  { "warning", "pwrite",  "XHamster.Warning",
                         "write error on pipe",               0 },
  /* E_PBROKEN  -8 */  { "warning", "pbroken", "XHamster.Warning",
                         "broken pipe",                       0 },
  /* E_FORK     -9 */  { "warning", "fork",    "XHamster.Warning",
                         "cannot fork child process",         0 },
  /* E_EXEC    -10 */  { "warning", "exec",    "XHamster.Warning",
                         "cannot execute program:\n%s",       1 },
  /* E_INIT    -11 */  { "error",   "init",    "XHamster.Error",
                         "initialization failed",             0 },
  /* E_WIDGET  -12 */  { "error",   "widget",  "XHamster.Error",
                         "widget creation failed",            0 },
  /* E_COLOR   -13 */  { "error",   "color",   "XHamster.Error",
                         "color allocation failed",           0 },
  /* E_GRAPH   -14 */  { "error",   "gc",      "XHamster.Error",
                         "cannot create graphics context",    0 },
  /* E_DIALOG  -15 */  { "warning", "dialog",  "XHamster.Warning",
                         "cannot create dialog box",          0 },
  /* E_UNKNOWN -16 */  { "error", "unknown",   "XHamster.Error",
                         "unknown error",                     0 } };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static AppData appdata = {      /* application specific data */
  20, 15,                       /* initial maze extensions */
  0.75, 500, 8,                 /* wall prob., total corn, max. heap */
#ifdef DEMO
  99,                           /* inital hamster speed */
#else
  2.5,                          /* inital hamster speed */
#endif
  0, 0, 0, 0,                   /* score variables */
  NONAME,                       /* file name of maze */
  "bin/chamster" };             /* file name of executable */
static XtAppContext appctx;     /* X11 application context */
static Atom      wm_delwin;     /* delete window atom of window mgr. */
static Display   *display;      /* the display connected to */
static int       screen;        /* screen of the display */
static char      fmt[FMT_MAX];  /* format and buffer for messages */
static char      buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget    w_top;         /* top level widget */
static Widget    w_main;        /* main window widget */
static Widget    w_view;        /* viewport widget for maze */
static Widget    w_maze;        /* maze display widget */
static Window    window;        /* window of maze widget */
static MENU      *menu;         /* pull down menu */
static GC        gc;            /* graphics context for maze */

/* --- property sheets --- */
static PSHEET    *psh_size   = NULL;  /* 'Set Maze Size' dialog box */
static PSHEET    *psh_random = NULL;  /* 'Randomize Maze' dialog box */
static PSHEET    *psh_speed  = NULL;  /* 'Set H. Speed' dialog box */
static PSHEET    *psh_result = NULL;  /* 'Result' dialog box */

/* --- color variables --- */
static ULONG     black;         /* black  (walls) */
static ULONG     white;         /* white  (normal fields) */
static ULONG     grey;          /* grey   (start field) */
static ULONG     yellow;        /* yellow (corn heaps) */
static ULONG     orange;        /* orange (border of corn heaps) */
static ULONG     red;           /* red    (the hamster) */

/* --- child process and pipe variables --- */
static pid_t     pid = -1;      /* child process identifier */
static FILE      *p_in;         /* input  pipe (to be read from) */
static FILE      *p_out;        /* output pipe (to be written to) */
static XtInputId ip_pipe;       /* pipe input procedure identifier */

/* --- hamster variables --- */
static Pixmap    hms_bmap;      /* bitmap of hamster sprite */
static MAZE      *maze;         /* maze the hamster is in */
static HMSCB     hmscb[1] = {   /* hamster control blocks */
  { NULL, NULL, "\0"} };        /* (at most one hamster) */
static int       xoff, yoff;    /* offsets for drawing */
static int       running = 0;   /* whether hamster is running */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (XtClass(widget) != transientShellWidgetClass)
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void dcb_alert (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (XtClass(widget) != transientShellWidgetClass)
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
  if (running) result();        /* if hamster is running, show result */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs -code;          /* get error message data and */
  va_start(va, code);           /* list of variable arguments */
  for (argc = 0; argc < (Cardinal)msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  argc = (Cardinal)msg->argc;   /* get number of arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is 'error' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is 'warning' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_alert, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*--------------------------------------------------------------------*/

static double drand (void)
{                               /* --- compute random number in [0,1) */
  return rand() /(RAND_MAX +1.0);
}  /* drand() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void draw_field (int x, int y)
{                               /* --- draw maze field */
  short c, t;                   /* contents of maze field and buffer */
  short cw, ce, cs, cn;         /* contents of surrounding fields */

  c  = mz_getfld(maze, x,   y); /* get contents of */
  cw = mz_getfld(maze, x-1, y); /* field to draw */
  ce = mz_getfld(maze, x+1, y); /* and contents of */
  cs = mz_getfld(maze, x, y-1); /* surrounding fields */
  cn = mz_getfld(maze, x, y+1);
  x  = xoff +x *FIELDWD;        /* multiply coordinates with */
  y  = yoff -y *FIELDWD;        /* field width and add offsets */

  /* --- draw field interior --- */
  XSetForeground(display, gc, (c & MZ_HOME) ? grey : white);
  XFillRectangle(display, window, gc,
    x, y -FIELDWD+2, FIELDWD-1, FIELDWD-1);

  /* --- draw corners --- */
  t = (short)((c & (MZ_WEST|MZ_SOUTH))
            | (cs & MZ_WEST) | (cw & MZ_SOUTH));
  XSetForeground(display, gc, (t) ? black : white);
  XDrawPoint(display, window, gc, x-1,         y+1);
  t = (short)((c & (MZ_EAST|MZ_SOUTH))
            | (cs & MZ_EAST) | (ce & MZ_SOUTH));
  XSetForeground(display, gc, (t) ? black : white);
  XDrawPoint(display, window, gc, x+FIELDWD-1, y+1);
  t = (short)((c & (MZ_EAST|MZ_NORTH))
            | (cn & MZ_EAST) | (ce & MZ_NORTH));
  XSetForeground(display, gc, (t) ? black : white);
  XDrawPoint(display, window, gc, x+FIELDWD-1, y-FIELDWD+1);
  t = (short)((c & (MZ_WEST|MZ_NORTH))
            | (cn & MZ_WEST) | (cw & MZ_NORTH));
  XSetForeground(display, gc, (t) ? black : white);
  XDrawPoint(display, window, gc, x-1,         y-FIELDWD+1);

  /* --- draw walls --- */
  XSetForeground(display, gc, (c & MZ_SOUTH) ? black : white);
  XDrawLine(display, window, gc,
    x,           y+1,         x+FIELDWD-2, y+1);
  XSetForeground(display, gc, (c & MZ_EAST) ? black : white);
  XDrawLine(display, window, gc,
    x+FIELDWD-1, y,           x+FIELDWD-1, y-FIELDWD+2);
  XSetForeground(display, gc, (c & MZ_NORTH) ? black : white);
  XDrawLine(display, window, gc,
    x,           y-FIELDWD+1, x+FIELDWD-2, y-FIELDWD+1);
  XSetForeground(display, gc, (c & MZ_WEST) ? black : white);
  XDrawLine(display, window, gc,
    x-1,         y,           x-1,         y-FIELDWD+2);

  /* --- draw corn heap --- */
  c &= MZ_ITEMS;                /* get corn on the field */
  if (c > 0) {                  /* if there is corn on the field */
    if (c > 12) c = 12;         /* limit heap size */
    t = (short)((c >> 1) +(c & 1));
    c = (c >> 1);               /* calculate circle extensions */
    x += (FIELDWD/2)-1 -t;      /* compute coordinates and width */
    y -= (FIELDWD/2)-1 +c; c = (short)(c+t);
    XSetForeground(display, gc, yellow);
    XFillArc(display, window, gc, x, y,
            (unsigned)c, (unsigned)c, 0, 360*64);
    XSetForeground(display, gc, orange);
    XDrawArc(display, window, gc, x, y,
            (unsigned)c, (unsigned)c, 0, 360*64);
  }                             /* draw corn heap */
}  /* draw_field() */

/*--------------------------------------------------------------------*/

static void hms_disp (HAMSTER *hms, int cmd, int arg)
{                               /* --- hamster display function */
  int    id;                    /* hamster identifier */
  int    x, y, sx, sy;          /* coordinates of hamster */
  int    dir;                   /* direction of hamster */
  SPRITE *sprite;               /* sprite to control */
  int    cnt;                   /* number of steps to move */
  int    series;                /* sprite image series */

  if (appdata.speed >= MAXSPEED)
    return;                     /* if maximal speed set, abort */
  id     = hms_id(hms);         /* get hamster identifier */
  sprite = hmscb[id].sprite;    /* and associated sprite */
  switch (cmd) {                /* evaluate command */
    case HMS_CREATED:           /* --- if hamster was created */
      mz_getpos(maze, &sx, &sy);/* get start position and */
      hms_pos(hms, &x, &y);     /* current position of hamster */
      x += sx; y += sy;         /* and compute absolute position */
      x = xoff +x *FIELDWD +(FIELDWD-SPRITEWD)/2;
      y = yoff -y *FIELDWD -(FIELDWD-SPRITEWD)/2 -SPRITEWD +1;
      dir = hms_dir(hms);       /* get direction (line of sight) */
      switch (dir) {            /* evaluate direction */
        case HMS_EAST : series = 0; break;
        case HMS_NORTH: series = 1; break;
        case HMS_WEST : series = 2; break;
        case HMS_SOUTH: series = 3; break;
        default       : return; /* get sprite image series number */
      }
      spr_select(sprite, series, 0);
      spr_draw(sprite, x, y);   /* select and draw sprite image */

    case HMS_SCORED:            /* --- if score has changed */
      sprintf(buf, "%d", hms_score(hms));
      mn_info(menu, buf); break;  /* set new score */

    case HMS_MOVED:             /* --- if hamster was moved */
      dir = hms_dir(hms);       /* get direction (line of sight) */
      switch (dir) {            /* and evaluate it */
        case HMS_EAST : x =  2; y =  0; cnt = FIELDWD/2; break;
        case HMS_NORTH: x =  0; y = -2; cnt = FIELDWD/2; break;
        case HMS_WEST : x = -2; y =  0; cnt = FIELDWD/2; break;
        case HMS_SOUTH: x =  0; y =  2; cnt = FIELDWD/2; break;
        default       : return; /* get offsets to new field */
      }
      if      (appdata.speed >= 16) { x *= 8; y *= 8; cnt /= 8; }
      else if (appdata.speed >=  8) { x *= 4; y *= 4; cnt /= 4; }
      else if (appdata.speed >=  4) { x *= 2; y *= 2; cnt /= 2; }
      spr_move(sprite, x, y, 0, cnt,
               1.0F/(float)cnt/appdata.speed, reply, 0);
      break;                    /* move sprite */

    case HMS_TURNED:            /* --- if hamster was turned */
      dir = hms_dir(hms);       /* get direction (line of sight) */
      switch (dir) {            /* and evaluate it */
        case HMS_EAST : series = (arg == HMS_POS) ? 3 : 5; break;
        case HMS_NORTH: series = (arg == HMS_POS) ? 0 : 6; break;
        case HMS_WEST : series = (arg == HMS_POS) ? 1 : 7; break;
        case HMS_SOUTH: series = (arg == HMS_POS) ? 2 : 4; break;
        default       : return; /* get sprite image series number */
      }                         /* and select image series */
      if (appdata.speed >= 50) {/* if not to show turns */
        spr_select(sprite, series, 6);
        return;                 /* simply select the last image */
      }                         /* and abort the function */
      spr_select(sprite, series, 0);  /* select image series */
      if      (appdata.speed >= 16) { x = 6; cnt = 1; }
      else if (appdata.speed >=  4) { x = 3; cnt = 2; }
      else                          { x = 1; cnt = 6; }
      spr_move(sprite, 0, 0, x, cnt,
               1.0F/(float)cnt/appdata.speed, reply, 0);
      break;                    /* turn sprite */

    case HMS_FIELD:             /* --- if field contents has changed */
      mz_getpos(maze, &sx, &sy);/* get start position and */
      hms_pos(hms, &x, &y);     /* current position of hamster */
      x += sx; y += sy;         /* and compute absolute position */
      spr_undraw(sprite);       /* undraw sprite */
      draw_field(x, y);         /* update changed field */
      spr_draw(sprite, INT_MIN, INT_MIN);
      break;                    /* redraw sprite */

    case HMS_DELETED:           /* --- if hamster was deleted */
      spr_undraw(sprite); break;/* undraw sprite */

    default: break;             /* ignore everything else */
  }
  XFlush(display);              /* flush buffered drawing commands */
}  /* hms_disp() */

/*--------------------------------------------------------------------*/

static void resize (Dimension width, Dimension height)
{                               /* --- resize widgets */
  Dimension w, h;               /* width and height of top widget */

  width  = (Dimension)(width  *FIELDWD);
  height = (Dimension)(height *FIELDWD);
  xoff = 0; yoff = height-2;    /* compute extensions and set offsets */
  w = (Dimension)(mn_minwidth(menu) -1);      /* get minimal width */
  if (w > width) width = w;     /* compute the maze viewport width */
  h = (Dimension)(mn_height(menu) +height);
  XtVaSetValues(w_top,  XtNwidth,    width+1, XtNheight,    h+1,
                        XtNmaxWidth, width+1, XtNmaxHeight, h+1, NULL);
  XtVaSetValues(w_maze, XtNwidth, width-1, XtNheight, height-1, NULL);
  XtVaSetValues(w_view, XtNwidth, width-1, XtNheight, height-1, NULL);
  mn_resize(menu, (Dimension)(width-1)); /* set new widget extensions */
}  /* resize() */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Maze...' callback */
  int  i = FN_MAX -1;           /* loop variable */
  char *src, *dst;              /* to traverse file names */
  FILE *file;                   /* file to read from */
  MAZE *tmp;                    /* temporary buffer for loaded maze */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if 'cancel' button pressed, abort */
  file = fopen(fname, "r");     /* open input file */
  if (!file) { error(E_FOPEN, fname); return; }
  tmp = mz_load(file);          /* load maze from file */
  fclose(file);                 /* close input file */
  if (!tmp) { error(E_FREAD, fname); return; }
  src = fname; dst = appdata.fn_maze;
  while (*src && (--i >= 0)) *dst++ = *src++;
  *dst = '\0';                  /* copy file name to buffer */
  mz_delete(maze); maze = tmp;  /* set new maze and draw it */
  mz_exts(maze, &appdata.xext, &appdata.yext);
  resize((Dimension)appdata.xext, (Dimension)appdata.yext);
  redraw(NULL, NULL, NULL, NULL);
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save Maze...' callback */
  int  i = FN_MAX -1;           /* loop variable */
  char *src, *dst;              /* to traverse file names */
  FILE *file;                   /* file to write to */
  int  res;                     /* result of function mz_save */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if 'cancel' button pressed, abort */
  file = fopen(fname, "w");     /* open output file */
  if (!file) { error(E_FOPEN, fname); return; }
  res = mz_save(maze, file);    /* save current maze to file */
  if ((fclose(file) != 0) || (res != 0))
    error(E_FWRITE, fname);     /* check for write error */
  src = fname; dst = appdata.fn_maze;
  while (*src && (--i >= 0)) *dst++ = *src++;
  *dst = '\0';                  /* copy file name to buffer */
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_select (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Select Hamster...' callback */
  int  i = FN_MAX -1;           /* loop variable */
  char *src, *dst;              /* to traverse file names */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if 'cancel' button pressed, abort */
  src = fname; dst = appdata.fn_hms;
  while (*src && (--i >= 0)) *dst++ = *src++;
  *dst = '\0';                  /* copy file name to buffer */
}  /* dcb_select() */

/*--------------------------------------------------------------------*/

static void dcb_size (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Set Maze Size...' callback */
  MAZE *tmp;                    /* temporary buffer for maze */
  int  xext, yext;              /* maze extensions */

  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
  if (!ok) return;              /* if 'cancel' button pressed, abort */
  xext = atoi(psh_getval(psh_size, "xExt"));  /* get x- and */
  yext = atoi(psh_getval(psh_size, "yExt"));  /* y-extension */
  if (xext < HMS_MINXEXT) xext = HMS_MINXEXT;
  if (xext > HMS_MAXXEXT) xext = HMS_MAXXEXT;
  if (yext < HMS_MINYEXT) yext = HMS_MINYEXT;
  if (yext > HMS_MAXYEXT) yext = HMS_MAXYEXT;
  tmp = mz_create(xext, yext);  /* get and check inputs */
  if (!tmp) return;             /* and try to create new maze */
  mz_delete(maze); maze = tmp;  /* if successful, delete old maze, */
  appdata.xext = xext;          /* set new maze, and */
  appdata.yext = yext;          /* note new extensions */
  xoff = 0; yoff = yext *FIELDWD -2;
  resize((Dimension)xext, (Dimension)yext);
  redraw(NULL,NULL,NULL,NULL);  /* set offsets, resize window, */
}  /* dcb_size() */             /* and redraw maze */

/*--------------------------------------------------------------------*/

static void dcb_random (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Randomize Maze...' callback */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
  if (!ok) return;              /* if 'cancel' button pressed, abort */
  appdata.wallprob = (float)(0.01
                   * atof(psh_getval(psh_random, "wallProb")));
  appdata.total    = atoi(psh_getval(psh_random, "total"));
  appdata.maxheap  = atoi(psh_getval(psh_random, "maxHeap"));
  if (appdata.wallprob < 0.0)     appdata.wallprob = 0.0F;
  if (appdata.wallprob > 1.0)     appdata.wallprob = 1.0F;
  if (appdata.total    < 0)       appdata.total    = 0;
  if (appdata.maxheap  < 1)       appdata.maxheap  = 1;
  if (appdata.maxheap  > MAXHEAP) appdata.maxheap  = MAXHEAP;
  mz_init(maze, drand,   appdata.wallprob,
          appdata.total, appdata.maxheap);
  redraw(NULL,NULL,NULL,NULL);  /* reinitialize and redraw maze */
}  /* dcb_random() */

/*--------------------------------------------------------------------*/

static void dcb_speed (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Set Hamster Speed' callback */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
  if (!ok) return;              /* if 'cancel' button pressed, abort */
  if (appdata.speed >= MAXSPEED)     /* if run without display, */
    redraw(NULL, NULL, NULL, NULL);  /* redraw maze */
  appdata.speed = (float)atof(psh_getval(psh_speed, "speed"));
  if (appdata.speed > MAXSPEED) appdata.speed = (float)MAXSPEED;
  if (appdata.speed < MINSPEED) appdata.speed = (float)MINSPEED;
}  /* dcb_speed() */

/*--------------------------------------------------------------------*/

static void dcb_result (Widget widget, XtPointer client, XtPointer call)
{                               /* --- 'Result' callback */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_result() */

/*----------------------------------------------------------------------
  Result Dialog Box Function
----------------------------------------------------------------------*/

static void result (void)
{                               /* --- 'Result' dialog box */
  if (!running) return;         /* if hamster is not running, abort */
  if (!psh_result) {             /* if dialog box is not created */
    psh_result = psh_create("resultDlg", w_top, 0);
    if (!psh_result) return;     /* create a property sheet */
    if ((psh_addline(psh_result, "corncntLabel")         != 0)
    ||  (psh_additem(psh_result, "corncnt", PSH_TEXT, 0) != 0)
    ||  (psh_addline(psh_result, "movecntLabel")         != 0)
    ||  (psh_additem(psh_result, "movecnt", PSH_TEXT, 0) != 0)
    ||  (psh_addline(psh_result, "crshcntLabel")         != 0)
    ||  (psh_additem(psh_result, "crshcnt", PSH_TEXT, 0) != 0)
    ||  (psh_addline(psh_result, "scoreLabel")           != 0)
    ||  (psh_additem(psh_result, "score",   PSH_TEXT, 0) != 0)) {
      psh_delete(psh_result); error(E_WIDGET); }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", appdata.corncnt);
  psh_setval(psh_result, "corncnt", buf);
  sprintf(buf, "%d", appdata.movecnt);
  psh_setval(psh_result, "movecnt", buf);
  sprintf(buf, "%d", appdata.crshcnt);
  psh_setval(psh_result, "crshcnt", buf);
  sprintf(buf, "%d", appdata.score);
  psh_setval(psh_result, "score",   buf);
  psh_handle(psh_result, dcb_result); /* set values and handle dialog */
  running = 0;                        /* clear running flag */
}  /* result() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_load   (Widget widget, XtPointer item, XtPointer call)
{ load(NULL, NULL, (String*)&item, &one); }

static void mcb_save   (Widget widget, XtPointer item, XtPointer call)
{ save(NULL, NULL, (String*)&item, &one); }

static void mcb_select (Widget widget, XtPointer item, XtPointer call)
{ select(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer item, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer item, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_size   (Widget widget, XtPointer item, XtPointer call)
{ size(NULL, NULL, NULL, NULL); }

static void mcb_random (Widget widget, XtPointer item, XtPointer call)
{ random(NULL, NULL, NULL, NULL); }

static void mcb_clear  (Widget widget, XtPointer item, XtPointer call)
{ clear(NULL, NULL, NULL, NULL); }

static void mcb_speed  (Widget widget, XtPointer item, XtPointer call)
{ speed(NULL, NULL, NULL, NULL); }

static void mcb_start  (Widget widget, XtPointer item, XtPointer call)
{ exec(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer item, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

static void mcb_menu   (Widget widget, XtPointer item, XtPointer call)
{ mn_info(menu, ""); }          /* clear info field in menu bar */

/*----------------------------------------------------------------------
  Child Process and Pipe Communication Functions
----------------------------------------------------------------------*/

static void stop (int errcode)
{                               /* --- stop hamster */
  HAMSTER *hamster;             /* the hamster and */
  SPRITE  *sprite;              /* its associated sprite */

  if (!running) return;         /* if hamster is not running, abort */
  kill(pid, SIGKILL);           /* kill child process and */
  wait(NULL);                   /* wait for termination */
  XtRemoveInput(ip_pipe);       /* remove pipe input procedure */
  fclose(p_in); fclose(p_out);  /* close both pipes and */
  hmscb[0].reply[0] = '\0';     /* clear buffered reply */
  hamster = hmscb[0].hamster;   /* get hamster */
  if (!hamster)                 /* if no hamster exists, */
    running = 0;                /* stop running at once */
  else {                        /* if a hamster exists */
    appdata.corncnt = hms_corncnt(hamster);
    appdata.movecnt = hms_movecnt(hamster);
    appdata.crshcnt = hms_crshcnt(hamster);
    appdata.score   = hms_score(hamster);
    if (errcode != E_NONE) appdata.score -= BDPENALTY;
    hms_delete(hamster); hmscb[0].hamster = NULL;
  }                             /* note score and delete hamster */
  sprite = hmscb[0].sprite;     /* get hamster sprite */
  if (sprite) {                 /* if a sprite exists, */
    spr_stop(sprite);           /* stop moving sprite, */
    spr_undraw(sprite);         /* then undraw sprite */
    spr_delete(sprite); hmscb[0].sprite = NULL;
  }                             /* delete sprite */
  if (appdata.speed >= MAXSPEED)     /* if run without display, */
    redraw(NULL, NULL, NULL, NULL);  /* redraw maze */
  mn_visible(menu, MN_ITEMS, 3);/* show/hide      menu items */
  mn_enable (menu, MN_ITEMS, 3);/* enable/disable menu items */
  mn_info(menu, "");            /* clear score */
  if (errcode != E_NONE)        /* show alert dialog box */
    error(errcode, appdata.fn_hms);
  else if (running)             /* if still running, */
    result();                   /* show result */
}  /* stop() */

/*--------------------------------------------------------------------*/

static void reply (Widget w, XtPointer client, XtPointer call)
{                               /* --- send reply to client program */
  if (!running) return;         /* if hamster is not running, abort */
  fprintf(p_out, "%s", hmscb[0].reply);     /* send reply to client */
  hmscb[0].reply[0] = '\0';          /* and mark reply as sent */
  if (fflush(p_out)) stop(E_PWRITE); /* flush output pipe buffer */
}  /* reply() */

/*--------------------------------------------------------------------*/

static void icb_pipe (XtPointer client, int *src, XtInputId *iid)
{                               /* --- pipe input callback function */
  HAMSTER *hamster;             /* hamster to control */
  char    cmd;                  /* command read */
  int     id = 0;               /* hamster identifier */
  int     arg;                  /* command argument */
  int     x, y;                 /* hamster position */

  if (!running) return;         /* if hamster is not running, abort */

  /* --- read command from pipe --- */
  if ( (fscanf(p_in, " %c", &cmd) != 1)
  ||  ((cmd != 'c')
  &&   (fscanf(p_in, " %d", &id)  != 1))
  ||  (((cmd == 't') || (cmd == 'l'))
  &&   (fscanf(p_in, " %d", &arg) != 1))
  ||   (getc(p_in) != '\n')) {  /* read next command from pipe */
    stop(feof(p_in) ? E_PBROKEN : E_PREAD); return; }

  if (((cmd != '@') && (id != 0))  /* if reply has not been sent yet, */
  ||  (hmscb[id].reply[0] != '\0')) {               /* stop execution */
    stop(feof(p_in) ? E_PBROKEN : E_PREAD); return; }
  hamster = hmscb[id].hamster;  /* get hamster to control */

  /* --- evaluate command --- */
  switch (cmd) {                /* evaluate command read */
    case '@':                   /* --- error in child process */
      stop((id) ? E_EXEC : E_POPEN);
      return;                   /* stop execution */

    case 'c':                   /* --- create hamster */
      if (hamster) {            /* if a hamster already exists */
        sprintf(hmscb[0].reply, "c -1\n");
        break;                  /* deny creation request and */
      }                         /* abort command evaluation */
      hmscb[0].sprite =         /* create a sprite */
        spr_create(w_maze, SPRITEWD, SPRITEWD, hms_bmap, red);
      if (!hmscb[0].sprite) { stop(E_NOMEM); return; }
      hmscb[0].hamster =        /* create a hamster */
      hamster = hms_create(maze, 0, HMS_EAST, hms_disp);
      if (!hamster) { stop(E_NOMEM); return; }
      hms_pos(hamster, &x, &y); /* get hamster position */
      sprintf(hmscb[0].reply, "c %d %d %d %d %d %d %d\n",
        hms_id(hamster), x, y, hms_dir(hamster), hms_look(hamster),
        hms_corn(hamster), hms_load(hamster));
      break;                    /* send hamster data */

    case 'd':                   /* --- delete hamster */
      fprintf(p_out, "d 0\n");  /* echo delete command, */
      fflush(p_out);            /* flush pipe buffer */
      stop(E_NONE); return;     /* and stop execution */

    case 'm':                   /* --- move hamster */
      hms_move(hamster);        /* move hamster forward */
      hms_pos(hamster, &x, &y); /* and get new position */
      sprintf(hmscb[id].reply, "m %d %d %d %d %d\n",
        hms_id(hamster), x, y, hms_look(hamster), hms_corn(hamster));
      break;                    /* reply with new hamster data */

    case 't':                   /* --- turn hamster */
      hms_turn(hamster, arg);   /* turn hamster around */
      sprintf(hmscb[id].reply, "t %d %d %d\n",
        hms_id(hamster), hms_dir(hamster), hms_look(hamster));
      break;                    /* reply with new hamster data */

    case 'l':                   /* --- take/drop corn */
      hms_take(hamster, arg);   /* take/drop some corn */
      sprintf(hmscb[id].reply, "l %d %d %d\n",
        hms_id(hamster), hms_corn(hamster), hms_load(hamster));
      break;                    /* send new hamster data */

    default:                    /* --- unknown command */
      stop(E_PREAD); return;    /* stop execution */
  }
  if (!spr_busy(hmscb[id].sprite)) /* if the sprite is not busy, */
    reply(NULL, NULL, NULL);       /* send reply directly */
}  /* icb_pipe() */

/*--------------------------------------------------------------------*/

static void start (void)
{                               /* --- start hamster */
  int p2c[2];                   /* pipe from parent to child */
  int c2p[2];                   /* pipe from child to parent */
  int fd_out, fd_in;            /* buffers for file descriptors */

  if (running) return;          /* if hamster is running, abort */
  mn_visible(menu, MN_ITEMS, 5);/* show/hide menu items */
  mn_enable (menu, MN_ITEMS, 4);/* enable/disable menu items */
  #ifdef DEMO                   /* for the demonstration version */
  { FILE *t;                    /* wait for a file 'demo' */
    do { t = fopen("demo", "r");/* to be created, before */
    } while (!t); fclose(t); }  /* starting the hamster */
  #endif
  if ((pipe(p2c) != 0) || (pipe(c2p) != 0)) {
    error(E_POPEN); return; }   /* create pipes and */
  pid = fork();                 /* fork child process */
  if (pid == -1) { error(E_FORK); return; }

  /* --- child process --- */
  if (pid ==  0) {              /* if this is the child process */
    close(STDOUT_FILENO);       /* connect first  pipe to stdout */
    fd_out = dup2(c2p[1], STDOUT_FILENO);
    close(STDIN_FILENO);        /* connect second pipe to stdin */
    fd_in  = dup2(p2c[0], STDIN_FILENO);
    close(c2p[0]); close(c2p[1]);  /* close pipe handles */
    close(p2c[0]); close(p2c[1]);  /* and check for success */
    if ((fd_out != STDOUT_FILENO) || (fd_in != STDIN_FILENO)) {
      printf("@ 0\n"); fflush(stdout); _exit(-1); }
    execl(appdata.fn_hms, appdata.fn_hms, NULL);
    printf("@ 1\n"); fflush(stdout); _exit(-1);
  }                             /* execute child program */

  /* --- parent process --- */
  close(p2c[0]); close(c2p[1]); /* close child ends of pipes */
  p_in  = fdopen(c2p[0], "r");  /* if this is the parent process, */
  p_out = fdopen(p2c[1], "w");  /* open streams on both pipes */
  if (!p_in || !p_out) {        /* and check for success */
    close(p2c[1]); close(c2p[0]);
    kill(pid, SIGKILL); error(E_POPEN);
    return;                     /* if an error occurred, */
  }                             /* close pipes and terminate child */
  ip_pipe = XtAppAddInput(appctx, c2p[0], (XtPointer)XtInputReadMask,
              icb_pipe, NULL);  /* install pipe read procedure */
  running = 1;                  /* set running flag and resize window */
  resize((Dimension)appdata.xext, (Dimension)appdata.yext);
}  /* start() */

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > (Re)Load Maze... */
  CCHAR *t;                     /* buffer for dialog box title */

  if (running || (*c != 1))     /* if running or wrong number */
    return;                     /* of arguments, abort function */
  if (((*s)[0] == '0')          /* if to reload maze */
  &&  (strcmp(appdata.fn_maze, NONAME) != 0))
    dcb_load(NULL, (void*)1, appdata.fn_maze);
  else {                        /* if to load new maze */
    mn_enable(menu, MN_MENU, 0);/* disable menu bar */
    XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
      XtNlabel, &t, NULL);      /* get menu item text */
    if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
      error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
  }                             /* get maze file name */
}  /* load() */

/*--------------------------------------------------------------------*/

static void save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save Maze... */
  CCHAR *t;                     /* buffer for dialog box title */

  if (running || (*c != 1))     /* if running or wrong number */
    return;                     /* of arguments, abort function */
  if (((*s)[0] == '0')          /* if to save maze */
  &&  (strcmp(appdata.fn_maze, NONAME) != 0))
    dcb_save(NULL, (void*)1, appdata.fn_maze);
  else {                        /* if to save under a new name */
    mn_enable(menu, MN_MENU, 0);/* disable menu bar */
    XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
      XtNlabel, &t, NULL);      /* get menu item text */
    if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
      error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
  }                             /* get maze file name */
}  /* save() */

/*--------------------------------------------------------------------*/

static void select (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Select Hamster... */
  CCHAR *t;                     /* buffer for dialog box title */

  if (running) return;          /* get name of hamster file to load */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*select"),
    XtNlabel, &t, NULL);        /* get menu item text */
  if (fselect(w_top, t, NULL, NULL, dcb_select) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* select() */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  if (running) stop(E_NONE);    /* if hamster is running, stop it */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  XRectangle rc;                /* clipping rectangle */
  int xext, yext;               /* maze extension */
  int x, y;                     /* loop variables */
  int t;                        /* result of spr_undraw */

  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  if (e && (e->type == Expose)){/* if called by Xt, */
    rc.x      = (short)e->xexpose.x; /* build and set */
    rc.y      = (short)e->xexpose.y; /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted);
  }
  t = (hmscb[0].sprite)         /* undraw sprite */
    ? spr_undraw(hmscb[0].sprite) : 1;
  mz_exts(maze, &xext, &yext);  /* get maze extensions */
  for (x = xext; --x >= 0; )    /* traverse columns */
    for (y = yext; --y >= 0; )  /* traverse lines of column */
      draw_field(x, y);         /* and draw field */
  if (t == 0)                   /* if sprite was drawn, redraw it */
    spr_draw(hmscb[0].sprite, INT_MIN, INT_MIN);
  if (!running)                 /* if not running, */
    mn_info(menu, "");          /* clear the score */
  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void size (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Set Maze Size... */
  char buf[16];                 /* buffer to format values */

  if (running) return;          /* if a hamster is running, abort */
  if (!psh_size) {              /* if dialog box is not created */
    psh_size = psh_create("sizeDlg", w_top, 1);
    if (!psh_size) return;      /* create a property sheet */
    if ((psh_addline(psh_size, "xExtLabel")         != 0)
    ||  (psh_additem(psh_size, "xExt", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_size, "yExtLabel")         != 0)
    ||  (psh_additem(psh_size, "yExt", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_size); error(E_WIDGET); }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", appdata.xext); psh_setval(psh_size, "xExt", buf);
  sprintf(buf, "%d", appdata.yext); psh_setval(psh_size, "yExt", buf);
  psh_handle(psh_size, dcb_size);  /* set values and handle dialog */
}  /* size() */

/*--------------------------------------------------------------------*/

static void random (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Randomize Maze... */
  char buf[16];                 /* buffer to format values */

  if (running) return;          /* if a hamster is running, abort */
  if (!psh_random) {            /* if dialog box is not created */
    psh_random = psh_create("randomDlg", w_top, 1);
    if (!psh_random) return;    /* create a property sheet */
    if ((psh_addline(psh_random, "wallProbLabel")         != 0)
    ||  (psh_additem(psh_random, "wallProb", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_random, "totalLabel")            != 0)
    ||  (psh_additem(psh_random, "total",    PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_random, "maxHeapLabel")          != 0)
    ||  (psh_additem(psh_random, "maxHeap",  PSH_EDIT, 0) != 0)) {
      psh_delete(psh_random); error(E_WIDGET); }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", (int)(appdata.wallprob *100 +0.5));
  psh_setval(psh_random, "wallProb", buf);
  sprintf(buf, "%d", appdata.total);
  psh_setval(psh_random, "total", buf);
  sprintf(buf, "%d", appdata.maxheap);
  psh_setval(psh_random, "maxHeap", buf);
  psh_handle(psh_random, dcb_random); /* set values and handle dialog */
}  /* random() */

/*--------------------------------------------------------------------*/

static void speed (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Set Hamster Speed... */
  char buf[16];                 /* buffer to format values */

  if (!psh_speed) {             /* if dialog box is not created */
    psh_speed = psh_create("speedDlg", w_top, 1);
    if (!psh_speed) return;     /* create a property sheet */
    if ((psh_addline(psh_speed, "speedLabel")         != 0)
    ||  (psh_additem(psh_speed, "speed", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_speed); error(E_WIDGET); }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%g", appdata.speed);
  psh_setval(psh_speed, "speed", buf);
  psh_handle(psh_speed, dcb_speed);  /* set values and handle dialog */
}  /* speed() */

/*--------------------------------------------------------------------*/

static void clear (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Clear Maze */
  if (running) return;          /* if a hamster is running, abort */
  mz_init(maze, 0, 0.0, 0, 0);  /* clear and redraw maze */
  redraw(NULL, NULL, NULL, NULL);
}  /* clear() */

/*--------------------------------------------------------------------*/

static void exec (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Start/Stop Hamster */
  if (running) stop(E_NONE);    /* if hamster is running, stop it, */
  else         start();         /* otherwise start hamster */
}  /* exec() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 1, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* about() */

/*--------------------------------------------------------------------*/

static void wall (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- toggle wall/remove walls */
  int   x, y, dx, dy;           /* coordinates and distances */
  short t;                      /* temporary buffer */

  if (running || (*c != 1)      /* if running or wrong arguments */
  ||  !XtIsSensitive(mn_menubar(menu))) /* or menu bar disabled, */
    return;                     /* abort function */
  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  x = e->xbutton.x -xoff;       /* compute coordinates relative */
  y = yoff -e->xbutton.y;       /* to lower left maze corner */
  if (strcmp(*s, "0") == 0) {   /* if to remove all walls, */
    x /= FIELDWD; y /= FIELDWD; /* clear all wall flags */
    if (x >= appdata.xext) return;
    mz_setfld(maze, x, y, MZ_WALLS, 0); }
  else if (strcmp(*s, "1") == 0) {
    dx = x % FIELDWD;           /* compute coordinates relative */
    dy = y % FIELDWD; t = 0;    /* to lower left field corner */
    if (dx >= FIELDWD/2) { t |= MZ_EAST;  dx = FIELDWD-1 -dx; }
    else                 { t |= MZ_WEST;  }
    if (dy >= FIELDWD/2) { t |= MZ_NORTH; dy = FIELDWD-1 -dy; }
    else                 { t |= MZ_SOUTH; }
    t = (short)(t & ((dx <= dy) ? MZ_WEST|MZ_EAST : MZ_SOUTH|MZ_NORTH));
    x /= FIELDWD; y /= FIELDWD; /* determine wall flag */
    if (x >= appdata.xext) return;
    t ^= mz_getfld(maze, x, y); /* get field contents */
    mz_setfld(maze, x, y, MZ_WALLS, t);
  }                             /* toggle wall flag */
  draw_field(x, y);             /* redraw changed field */
  mn_info(menu, "");            /* clear info field in menu bar */
}  /* wall() */

/*--------------------------------------------------------------------*/

static void heap (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- change corn heap size */
  int   x, y, sx, sy;           /* maze coordinates */
  short heap, chg;              /* size of corn heap and change */

  if (running || (*c != 1)      /* if running or wrong arguments */
  ||  !XtIsSensitive(mn_menubar(menu))) /* or menu bar disabled, */
    return;                     /* abort function */
  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  x = (e->xbutton.x -xoff) /FIELDWD; /* compute maze */
  y = (yoff -e->xbutton.y) /FIELDWD; /* coordinates */
  mz_getpos(maze, &sx, &sy);    /* get coord. of start position */
  if ((x == sx) && (y == sy))   /* do not allow to place */
    return;                     /* corn on the home field */
  heap = mz_getfld(maze, x, y) & MZ_ITEMS;
  chg  = (short)atoi(*s);       /* get heap size and change */
  heap = (short)((chg == 0) ? 0
       : ((heap +chg +(MAXHEAP+1)) %(MAXHEAP+1)));
  mz_setfld(maze, x, y, MZ_ITEMS, heap);
  draw_field(x, y);             /* set new heap size and redraw field */
  sprintf(buf, "%d", heap);     /* show new heap size */
  mn_info(menu, buf);           /* in info field in menu bar */
}  /* heap() */

/*--------------------------------------------------------------------*/

static void position (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- change start position */
  int ox, oy, nx, ny;           /* old and new start position */

  if (running || (*c != 0)      /* if running or wrong arguments */
  ||  !XtIsSensitive(mn_menubar(menu))) /* or menu bar disabled, */
    return;                     /* abort function */
  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  nx = (e->xbutton.x -xoff) /FIELDWD; /* compute maze */
  ny = (yoff -e->xbutton.y) /FIELDWD; /* coordinates */
  mz_getpos(maze, &ox, &oy);    /* get old start position */
  mz_setpos(maze, nx,   ny);    /* and then set new */
  draw_field(ox, oy);           /* redraw fields of old */
  draw_field(nx, ny);           /* and new start position */
  mn_info(menu, "");            /* clear info field in menu bar */
}  /* position() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg       args[8];            /* widget arguments */
  Cardinal  n;                  /* argument counter */
  Dimension wd, ht, m;          /* width and height of window */
  Colormap  colmap;             /* color map of display */
  XColor    xc[5];              /* buffer for colors */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create a maze --- */
  srand((unsigned int)time(NULL));
  maze = mz_create(appdata.xext, appdata.yext);
  if (!maze) error(E_NOMEM);    /* create a maze */
  wd = (Dimension)(appdata.xext *FIELDWD); /* compute initial */
  ht = (Dimension)(appdata.yext *FIELDWD); /* widget size */

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "XHamster",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET);

  /* --- create menu bar --- */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",    0, mcb_menu,   (void*)0) != 0)
  ||  (mn_additem (menu, "reload",  1, mcb_load,   "0")      != 0)
  ||  (mn_additem (menu, "load",    1, mcb_load,   "1")      != 0)
  ||  (mn_additem (menu, NULL,      0, 0,          (void*)0) != 0)
  ||  (mn_additem (menu, "save",    1, mcb_save,   "0")      != 0)
  ||  (mn_additem (menu, "saveas",  1, mcb_save,   "1")      != 0)
  ||  (mn_additem (menu, NULL,      0, 0,          (void*)0) != 0)
  ||  (mn_additem (menu, "select",  1, mcb_select, (void*)0) != 0)
  ||  (mn_additem (menu, NULL,      0, 0,          (void*)0) != 0)
  ||  (mn_additem (menu, "quit",    0, mcb_quit,   (void*)0) != 0)
  ||  (mn_addtitle(menu, "actions", 0, mcb_menu,   (void*)0) != 0)
  ||  (mn_additem (menu, "redraw",  0, mcb_redraw, (void*)0) != 0)
  ||  (mn_additem (menu, NULL,      0, 0,          (void*)0) != 0)
  ||  (mn_additem (menu, "size",    1, mcb_size,   (void*)0) != 0)
  ||  (mn_additem (menu, "random",  1, mcb_random, (void*)0) != 0)
  ||  (mn_additem (menu, "clear",   1, mcb_clear,  (void*)0) != 0)
  ||  (mn_additem (menu, NULL,      0, 0,          (void*)0) != 0)
  ||  (mn_additem (menu, "speed",   0, mcb_speed,  (void*)0) != 0)
  ||  (mn_additem (menu, "start",   2, mcb_start,  (void*)0) != 0)
  ||  (mn_additem (menu, "stop",    4, mcb_start,  (void*)0) != 0)
  ||  (mn_addtitle(menu, "help",    0, mcb_menu,   (void*)0) != 0)
  ||  (mn_additem (menu, "about",   0, mcb_about,  (void*)0) != 0)
  ||  (mn_addtitle(menu, NULL,      0, 0,          (void*)0) != 0))
    error(E_WIDGET);
  mn_visible(menu, MN_ITEMS, 3);/* show/hide      menu items */
  mn_enable (menu, MN_ITEMS, 3);/* enable/disable menu items */

  /* --- create maze view --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);   n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);    n++;
  XtSetArg(args[n], XtNright,     XtChainRight);  n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom); n++;
  XtSetArg(args[n], XtNresizable, True);          n++;
  w_view = XtCreateManagedWidget("view",
             viewportWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */

  /* --- create maze widget --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);           n++;
  XtSetArg(args[n], XtNleft,        XtChainLeft); n++;
  XtSetArg(args[n], XtNtop,         XtChainTop);  n++;
  XtSetArg(args[n], XtNright,       XtChainLeft); n++;
  XtSetArg(args[n], XtNbottom,      XtChainTop);  n++;
  XtSetArg(args[n], XtNresizable,   True);        n++;
  w_maze = XtCreateManagedWidget("maze",
             coreWidgetClass, w_view, args, n);
  if (!w_maze) error(E_WIDGET); /* create box for maze and */
  xoff = 0; yoff = ht-2;        /* compute offsets for drawing */
  XtVaSetValues(w_maze, XtNwidth, wd-1, XtNheight, ht-1, NULL);
  XtOverrideTranslations(w_maze,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));
  wd = HMS_MINXEXT*FIELDWD+1;   /* compute minimal width and height */
  ht = (Dimension)(HMS_MINYEXT*FIELDWD+1 +mn_height(menu));
  m  = (Dimension)mn_minwidth(menu); /* get minimal menu width */
  if (m > wd) wd = m;           /* and adapt the minimal width */
  n  = 0;                       /* inhibit undersizing */
  XtSetArg(args[n], XtNminWidth,  wd); n++;
  XtSetArg(args[n], XtNminHeight, ht); n++;
  XtSetValues(w_top, args, n);  /* set minimal width and height */

  /* --- install actions and accelerators --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_maze);   /* and drawable for window */
  resize((Dimension)appdata.xext, (Dimension)appdata.yext);

  if (db_alert(w_top, dcb_alert, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- allocate colors --- */
  screen  = DefaultScreen(display);
  colmap  = DefaultColormap(display, screen);
  white   = WhitePixel(display, screen);
  black   = BlackPixel(display, screen);
  if ((XAllocNamedColor(display, colmap, "grey",   xc+1, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "yellow", xc+2, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "orange", xc+3, xc) == 0)
  ||  (XAllocNamedColor(display, colmap, "red",    xc+4, xc) == 0))
    error(E_COLOR);             /* allocate colors and */
  grey   = xc[1].pixel;         /* note their pixel values */
  yellow = xc[2].pixel;
  orange = xc[3].pixel;
  red    = xc[4].pixel;

  /* --- create hamster sprite bitmap --- */
  hms_bmap = XCreateBitmapFromData(display, window,
             (char*)hms_bits, hms_width, hms_height);
  if (hms_bmap == None) error(E_NOMEM);

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
